"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = require("http");
const https_1 = require("https");
const url_1 = require("url");
const node_1 = require("@reactive-js/node");
const option_1 = require("@reactive-js/option");
const observable_1 = require("@reactive-js/observable");
const disposable_1 = require("@reactive-js/disposable");
const http_2 = require("@reactive-js/http");
const async_enumerable_1 = require("@reactive-js/async-enumerable");
const pipe_1 = require("@reactive-js/pipe");
const httpRequest_1 = require("./httpRequest");
class RequestBody extends disposable_1.AbstractDisposable {
    constructor(req) {
        super();
        this.req = req;
        this.consumed = false;
        this.add(_ => {
            req.removeAllListeners();
            req.once("error", () => { });
            req.once("close", () => {
                req.removeAllListeners();
            });
            req.destroy();
        });
        const onError = (cause) => {
            this.dispose({ cause });
        };
        req.on("error", onError);
    }
    enumerateAsync(scheduler, replayCount) {
        if (this.consumed) {
            throw new Error("Request body already consumed");
        }
        this.consumed = true;
        const sink = node_1.createBufferStreamSinkFromWritable(() => this.req, false)
            .enumerateAsync(scheduler, replayCount)
            .add(this);
        this.add(sink);
        return sink;
    }
}
class ResponseBody extends disposable_1.AbstractDisposable {
    constructor(resp) {
        super();
        this.resp = resp;
        this.consumed = false;
        this.add(_ => {
            resp.removeAllListeners();
            resp.destroy();
        });
        const onError = (cause) => {
            this.dispose({ cause });
        };
        resp.on("error", onError);
    }
    enumerateAsync(scheduler, replayCount) {
        if (this.consumed) {
            throw new Error("Response body already consumed");
        }
        this.consumed = true;
        const stream = node_1.createBufferStreamFromReadable(() => this.resp)
            .enumerateAsync(scheduler, replayCount)
            .add(this);
        this.add(stream);
        return stream;
    }
}
exports.createHttpClient = (options = {}) => request => {
    return observable_1.createObservable(subscriber => {
        const { method, uri } = request;
        const url = uri instanceof url_1.URL ? uri : new url_1.URL(uri.toString());
        const headers = http_2.httpRequestToUntypedHeaders(request);
        const nodeRequestOptions = Object.assign(Object.assign({}, options), { headers,
            method });
        const req = uri.protocol === "https:"
            ? https_1.request(url, nodeRequestOptions)
            : uri.protocol === "http:"
                ? http_1.request(url, nodeRequestOptions)
                : (() => {
                    throw new Error();
                })();
        const reqBody = new RequestBody(req);
        const sinkSubscription = disposable_1.createSerialDisposable().add(reqBody);
        subscriber.add(sinkSubscription);
        const onResponse = (resp) => {
            var _a;
            sinkSubscription.dispose();
            const body = new ResponseBody(resp);
            subscriber.add(body);
            body.add(subscriber);
            const response = http_2.parseHttpResponseFromHeaders((_a = resp.statusCode) !== null && _a !== void 0 ? _a : -1, resp.headers, body);
            subscriber.dispatch({
                type: 4,
                response,
            });
        };
        req.on("response", onResponse);
        const content = pipe_1.pipe(request.body, async_enumerable_1.lift(ev => observable_1.createObservable(streamEventSubscriber => {
            const publishedEvents = pipe_1.pipe(ev, observable_1.publish(streamEventSubscriber));
            const progressSubscription = pipe_1.pipe(publishedEvents, observable_1.scan(([incr, count], ev) => ev.type === 1
                ? [ev.data.length, count + incr]
                : [-1, count + incr], () => [0, 0]), observable_1.onNotify(([incr, count]) => {
                const ev = incr < 0
                    ? { type: 3 }
                    : count > 0
                        ? {
                            type: 2,
                            count,
                        }
                        : option_1.none;
                if (option_1.isSome(ev)) {
                    subscriber.notify(ev);
                }
            }), observable_1.subscribe(streamEventSubscriber));
            streamEventSubscriber.add(publishedEvents).add(progressSubscription);
            publishedEvents.subscribe(streamEventSubscriber);
        })));
        subscriber.dispatch({ type: 1 });
        const onContinue = () => {
            sinkSubscription.inner = pipe_1.pipe(async_enumerable_1.sink(content, reqBody), observable_1.subscribe(subscriber));
        };
        if (request.expectContinue) {
            req.on("continue", onContinue);
        }
        else {
            onContinue();
        }
    });
};
const redirectCodes = [
    301,
    302,
    303,
    307,
    308,
];
exports.withDefaultBehaviors = (options) => (httpClient) => {
    const sendRequest = request => pipe_1.pipe(observable_1.ofValue(request), observable_1.map(httpRequest_1.encodeHttpRequest(options)), observable_1.switchMap(httpClient), observable_1.concatMap(status => {
        var _a, _b;
        if (status.type === 4) {
            const { response } = status;
            const { location, preferences, statusCode } = response;
            const acceptedEncodings = (_a = preferences === null || preferences === void 0 ? void 0 : preferences.acceptedEncodings) !== null && _a !== void 0 ? _a : [];
            const shouldRedirect = redirectCodes.includes(statusCode) &&
                option_1.isSome(location) &&
                ((_b = request === null || request === void 0 ? void 0 : request.maxRedirects) !== null && _b !== void 0 ? _b : 10) > 0;
            const newRequest = shouldRedirect
                ? http_2.createRedirectHttpRequest(request, response)
                : statusCode === 417
                    ? Object.assign(Object.assign({}, request), { expectContinue: false }) : statusCode === 415 &&
                    acceptedEncodings.length > 0
                    ? Object.assign(Object.assign({}, request), { acceptedEncodings }) : request;
            if (request !== newRequest) {
                response.body.dispose();
                return sendRequest(newRequest);
            }
        }
        return observable_1.ofValue(status);
    }));
    return sendRequest;
};
//# sourceMappingURL=httpClient.js.map