import { request as httpRequest } from "http";
import { request as httpsRequest } from "https";
import { URL } from "url";
import { createBufferStreamFromReadable, createBufferStreamSinkFromWritable, } from "@reactive-js/node";
import { none, isSome } from "@reactive-js/option";
import { createObservable, publish, subscribe, scan, onNotify, ofValue, map, concatMap, switchMap, } from "@reactive-js/observable";
import { AbstractDisposable, createSerialDisposable, } from "@reactive-js/disposable";
import { httpRequestToUntypedHeaders, parseHttpResponseFromHeaders, createRedirectHttpRequest, } from "@reactive-js/http";
import { lift, sink, } from "@reactive-js/async-enumerable";
import { pipe } from "@reactive-js/pipe";
import { encodeHttpRequest } from "./httpRequest";
class RequestBody extends AbstractDisposable {
    constructor(req) {
        super();
        this.req = req;
        this.consumed = false;
        this.add(_ => {
            req.removeAllListeners();
            req.once("error", () => { });
            req.once("close", () => {
                req.removeAllListeners();
            });
            req.destroy();
        });
        const onError = (cause) => {
            this.dispose({ cause });
        };
        req.on("error", onError);
    }
    enumerateAsync(scheduler, replayCount) {
        if (this.consumed) {
            throw new Error("Request body already consumed");
        }
        this.consumed = true;
        const sink = createBufferStreamSinkFromWritable(() => this.req, false)
            .enumerateAsync(scheduler, replayCount)
            .add(this);
        this.add(sink);
        return sink;
    }
}
class ResponseBody extends AbstractDisposable {
    constructor(resp) {
        super();
        this.resp = resp;
        this.consumed = false;
        this.add(_ => {
            resp.removeAllListeners();
            resp.destroy();
        });
        const onError = (cause) => {
            this.dispose({ cause });
        };
        resp.on("error", onError);
    }
    enumerateAsync(scheduler, replayCount) {
        if (this.consumed) {
            throw new Error("Response body already consumed");
        }
        this.consumed = true;
        const stream = createBufferStreamFromReadable(() => this.resp)
            .enumerateAsync(scheduler, replayCount)
            .add(this);
        this.add(stream);
        return stream;
    }
}
export const createHttpClient = (options = {}) => request => {
    return createObservable(subscriber => {
        const { method, uri } = request;
        const url = uri instanceof URL ? uri : new URL(uri.toString());
        const headers = httpRequestToUntypedHeaders(request);
        const nodeRequestOptions = Object.assign(Object.assign({}, options), { headers,
            method });
        const req = uri.protocol === "https:"
            ? httpsRequest(url, nodeRequestOptions)
            : uri.protocol === "http:"
                ? httpRequest(url, nodeRequestOptions)
                : (() => {
                    throw new Error();
                })();
        const reqBody = new RequestBody(req);
        const sinkSubscription = createSerialDisposable().add(reqBody);
        subscriber.add(sinkSubscription);
        const onResponse = (resp) => {
            var _a;
            sinkSubscription.dispose();
            const body = new ResponseBody(resp);
            subscriber.add(body);
            body.add(subscriber);
            const response = parseHttpResponseFromHeaders((_a = resp.statusCode) !== null && _a !== void 0 ? _a : -1, resp.headers, body);
            subscriber.dispatch({
                type: 4,
                response,
            });
        };
        req.on("response", onResponse);
        const content = pipe(request.body, lift(ev => createObservable(streamEventSubscriber => {
            const publishedEvents = pipe(ev, publish(streamEventSubscriber));
            const progressSubscription = pipe(publishedEvents, scan(([incr, count], ev) => ev.type === 1
                ? [ev.data.length, count + incr]
                : [-1, count + incr], () => [0, 0]), onNotify(([incr, count]) => {
                const ev = incr < 0
                    ? { type: 3 }
                    : count > 0
                        ? {
                            type: 2,
                            count,
                        }
                        : none;
                if (isSome(ev)) {
                    subscriber.notify(ev);
                }
            }), subscribe(streamEventSubscriber));
            streamEventSubscriber.add(publishedEvents).add(progressSubscription);
            publishedEvents.subscribe(streamEventSubscriber);
        })));
        subscriber.dispatch({ type: 1 });
        const onContinue = () => {
            sinkSubscription.inner = pipe(sink(content, reqBody), subscribe(subscriber));
        };
        if (request.expectContinue) {
            req.on("continue", onContinue);
        }
        else {
            onContinue();
        }
    });
};
const redirectCodes = [
    301,
    302,
    303,
    307,
    308,
];
export const withDefaultBehaviors = (options) => (httpClient) => {
    const sendRequest = request => pipe(ofValue(request), map(encodeHttpRequest(options)), switchMap(httpClient), concatMap(status => {
        var _a, _b;
        if (status.type === 4) {
            const { response } = status;
            const { location, preferences, statusCode } = response;
            const acceptedEncodings = (_a = preferences === null || preferences === void 0 ? void 0 : preferences.acceptedEncodings) !== null && _a !== void 0 ? _a : [];
            const shouldRedirect = redirectCodes.includes(statusCode) &&
                isSome(location) &&
                ((_b = request === null || request === void 0 ? void 0 : request.maxRedirects) !== null && _b !== void 0 ? _b : 10) > 0;
            const newRequest = shouldRedirect
                ? createRedirectHttpRequest(request, response)
                : statusCode === 417
                    ? Object.assign(Object.assign({}, request), { expectContinue: false }) : statusCode === 415 &&
                    acceptedEncodings.length > 0
                    ? Object.assign(Object.assign({}, request), { acceptedEncodings }) : request;
            if (request !== newRequest) {
                response.body.dispose();
                return sendRequest(newRequest);
            }
        }
        return ofValue(status);
    }));
    return sendRequest;
};
//# sourceMappingURL=httpClient.js.map